@@ -1,5 +1,5 @@
 /*
- * Copyright 2021-2022 Robert Tari <robert@tari.in>
+ * Copyright 2021-2023 Robert Tari <robert@tari.in>
  *
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 3, as published
@@ -21,6 +21,9 @@
 #include "languages.h"
 #include "keyboard.h"
 
+#define GREETER_BUS_NAME "org.ayatana.greeter"
+#define GREETER_BUS_PATH "/org/ayatana/greeter"
+
 enum
 {
     LAYOUT_CHANGED,
@@ -42,6 +45,10 @@ struct _KeyboardPrivate
     XklConfigRec *pConfigRec;
     GSList *lLayoutRec;
     GSList *lUsers;
+    GDBusConnection *pConnection;
+    guint nSubscription;
+    gchar *sUser;
+    gchar *sSystemLayout;
 };
 
 typedef KeyboardPrivate priv_t;
@@ -114,22 +121,54 @@ static void setAccountsService(Keyboard *pKeyboard)
     g_free(sPath);
 }
 
-static void getAccountsService(Keyboard *pKeyboard)
+static void getAccountsService(Keyboard *pKeyboard, ActUser *pUser)
 {
-    gboolean bChanged = FALSE;
-
-    for (GSList *lUser = pKeyboard->pPrivate->lUsers; lUser; lUser = lUser->next)
+    if (!pKeyboard->pPrivate->sUser)
     {
-        ActUser *pUser = lUser->data;
-        gboolean bIsUserLoaded = act_user_is_loaded(pUser);
+        GError *pError = NULL;
+        GVariant *pUser = g_dbus_connection_call_sync (pKeyboard->pPrivate->pConnection, GREETER_BUS_NAME, GREETER_BUS_PATH, GREETER_BUS_NAME, "GetUser", NULL, G_VARIANT_TYPE ("(s)"), G_DBUS_CALL_FLAGS_NONE, -1, NULL, &pError);
 
-        if (bIsUserLoaded)
+        if (pError)
         {
+            g_warning ("Failed calling GetUser, the greeter may not be ready yet: %s", pError->message);
+            g_error_free (pError);
+
+            return;
+        }
+
+        g_variant_get (pUser, "(s)", &pKeyboard->pPrivate->sUser);
+    }
+
+    gboolean bPrefix = g_str_has_prefix (pKeyboard->pPrivate->sUser, "*");
+
+    if (bPrefix)
+    {
+        if (pKeyboard->pPrivate->lLayoutRec)
+        {
+            g_slist_free_full (g_steal_pointer (&pKeyboard->pPrivate->lLayoutRec), g_free);
+        }
+
+        pKeyboard->pPrivate->lLayoutRec = g_slist_append (pKeyboard->pPrivate->lLayoutRec, g_strdup (pKeyboard->pPrivate->sSystemLayout));
+        g_timeout_add(500, (GSourceFunc)emitDelayedSignal, pKeyboard);
+    }
+    else
+    {
+        const gchar *sUser = act_user_get_user_name (pUser);
+        gboolean bSame = g_str_equal(pKeyboard->pPrivate->sUser, sUser);
+
+        if (bSame)
+        {
+            if (pKeyboard->pPrivate->lLayoutRec)
+            {
+                g_slist_free_full (g_steal_pointer (&pKeyboard->pPrivate->lLayoutRec), g_free);
+            }
+
             gint nUid = act_user_get_uid(pUser);
             gchar *sPath = g_strdup_printf("/org/freedesktop/Accounts/User%i", nUid);
             GDBusConnection *pConnection = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, NULL);
             GDBusProxy *pProxy = g_dbus_proxy_new_sync(pConnection, G_DBUS_PROXY_FLAGS_NONE, NULL, "org.freedesktop.Accounts", sPath, "org.freedesktop.DBus.Properties", NULL, NULL);
             GVariant *pValue = g_dbus_proxy_call_sync(pProxy, "Get", g_variant_new("(ss)", "org.ayatana.indicator.keyboard.AccountsService", "Layouts"), G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL);
+
             GVariant *pChild0 = g_variant_get_child_value(pValue, 0);
             GVariant *pChild1 = g_variant_get_child_value(pChild0, 0);
 
@@ -141,11 +180,14 @@ static void getAccountsService(Keyboard *pKeyboard)
 
             while (g_variant_iter_loop(pIter, "s", &sLayout))
             {
-                if (g_slist_find_custom(pKeyboard->pPrivate->lLayoutRec, sLayout, (GCompareFunc)g_strcmp0) == NULL)
-                {
-                    pKeyboard->pPrivate->lLayoutRec = g_slist_append(pKeyboard->pPrivate->lLayoutRec, g_strdup(sLayout));
-                    bChanged = TRUE;
-                }
+                pKeyboard->pPrivate->lLayoutRec = g_slist_append(pKeyboard->pPrivate->lLayoutRec, g_strdup(sLayout));
+            }
+
+            guint nLayouts = g_slist_length(pKeyboard->pPrivate->lLayoutRec);
+
+            if (!nLayouts)
+            {
+                pKeyboard->pPrivate->lLayoutRec = g_slist_append (pKeyboard->pPrivate->lLayoutRec, g_strdup (pKeyboard->pPrivate->sSystemLayout));
             }
 
             g_variant_iter_free(pIter);
@@ -154,12 +196,9 @@ static void getAccountsService(Keyboard *pKeyboard)
             g_variant_unref(pValue);
             g_object_unref(pConnection);
             g_free(sPath);
-        }
-    }
 
-    if (bChanged == TRUE)
-    {
-        g_timeout_add(500, (GSourceFunc)emitDelayedSignal, pKeyboard);
+            g_timeout_add(500, (GSourceFunc)emitDelayedSignal, pKeyboard);
+        }
     }
 }
 
@@ -169,7 +208,7 @@ static void onUserLoaded(Keyboard *pKeyboard, ActUser *pUser)
 
     if (bIsUserLoaded)
     {
-        getAccountsService(pKeyboard);
+        getAccountsService(pKeyboard, pUser);
         g_signal_handlers_disconnect_by_func(G_OBJECT(pUser), G_CALLBACK(onUserLoaded), pKeyboard);
     }
 }
@@ -182,7 +221,10 @@ static void onManagerLoaded(Keyboard *pKeyboard)
 
     if (bIsLoaded)
     {
-        pKeyboard->pPrivate->lUsers = act_user_manager_list_users(pManager);
+        if (!pKeyboard->pPrivate->lUsers)
+        {
+            pKeyboard->pPrivate->lUsers = act_user_manager_list_users(pManager);
+        }
 
         for (GSList *lUser = pKeyboard->pPrivate->lUsers; lUser; lUser = lUser->next)
         {
@@ -191,7 +233,7 @@ static void onManagerLoaded(Keyboard *pKeyboard)
 
             if (bIsUserLoaded)
             {
-                getAccountsService(pKeyboard);
+                getAccountsService(pKeyboard, pUser);
             }
             else
             {
@@ -317,14 +359,23 @@ void keyboard_AddSource(Keyboard *pKeyboard)
 
 guint keyboard_GetNumLayouts(Keyboard *pKeyboard)
 {
+    guint nLayouts = 0;
+
     if (isGreeter())
     {
-        return g_slist_length(pKeyboard->pPrivate->lLayoutRec);
+        nLayouts = g_slist_length(pKeyboard->pPrivate->lLayoutRec);
     }
     else
     {
-        return g_strv_length(pKeyboard->pPrivate->pConfigRec->layouts);
+        nLayouts = g_strv_length(pKeyboard->pPrivate->pConfigRec->layouts);
     }
+
+    return nLayouts;
+}
+
+guint keyboard_GetLayoutIndex (Keyboard *pKeyboard)
+{
+    return pKeyboard->pPrivate->nLayout;
 }
 
 void keyboard_GetLayout(Keyboard *pKeyboard, gint nLayout, gchar **pLanguage, gchar **pDescription)
@@ -380,34 +431,29 @@ void keyboard_SetLayout(Keyboard *pKeyboard, gint nLayout)
     }
     else
     {
-        // TODO
-        gchar *sCommand;
         gchar *sId = g_slist_nth_data(pKeyboard->pPrivate->lLayoutRec, nLayout);
-        gchar **lParamas = g_strsplit(sId, "+", -1);
-        guint nParams = g_strv_length(lParamas);
+        gchar **lParams = g_strsplit(sId, "+", -1);
+        guint nParams = g_strv_length(lParams);
+        gchar *sVariant = "";
 
-        if (nParams == 1)
+        if (nParams == 2)
         {
-            sCommand = g_strdup_printf("setxkbmap -layout %s", lParamas[0]);
-        }
-        else
-        {
-            sCommand = g_strdup_printf("setxkbmap -layout %s -variant %s", lParamas[0], lParamas[1]);
+            sVariant = lParams[1];
         }
 
-        g_strfreev(lParamas);
-
-        gchar *sOutput = NULL;
         GError *pError = NULL;
-        gboolean bResult = g_spawn_command_line_sync(sCommand, &sOutput, NULL, NULL, &pError);
-        g_free(sCommand);
+        GVariant *pLayout = g_variant_new ("(ss)", lParams[0], sVariant);
+        g_dbus_connection_call_sync (pKeyboard->pPrivate->pConnection, GREETER_BUS_NAME, GREETER_BUS_PATH, GREETER_BUS_NAME, "SetLayout", pLayout, NULL, G_DBUS_CALL_FLAGS_NONE, -1, NULL, &pError);
 
-        if (bResult == FALSE)
+        if (pError)
         {
-            g_message("COMMAND 01 %s: %s", sOutput, pError->message);
+            g_error ("Panic: Failed calling SetLayout: %s", pError->message);
+            g_error_free (pError);
 
             return;
         }
+
+        g_strfreev (lParams);
     }
 
     pKeyboard->pPrivate->nLayout = nLayout;
@@ -416,7 +462,22 @@ void keyboard_SetLayout(Keyboard *pKeyboard, gint nLayout)
 
 static void onDispose(GObject *pObject)
 {
-    Keyboard *self = G_KEYBOARD(pObject);
+    Keyboard *self = G_KEYBOARD (pObject);
+
+    if (self->pPrivate->nSubscription)
+    {
+        g_dbus_connection_signal_unsubscribe (self->pPrivate->pConnection, self->pPrivate->nSubscription);
+    }
+
+    if (self->pPrivate->lUsers)
+    {
+        g_slist_free(self->pPrivate->lUsers);
+    }
+
+    if (self->pPrivate->pConnection)
+    {
+        g_object_unref (self->pPrivate->pConnection);
+    }
 
     if (self->pPrivate->lLayouts)
     {
@@ -433,9 +494,9 @@ static void onDispose(GObject *pObject)
         g_slist_free_full(self->pPrivate->lLayoutRec, g_free);
     }
 
-    if (self->pPrivate->lUsers)
+    if (self->pPrivate->sSystemLayout)
     {
-        g_slist_free(self->pPrivate->lUsers);
+        g_free (self->pPrivate->sSystemLayout);
     }
 
     G_OBJECT_CLASS(keyboard_parent_class)->dispose(pObject);
@@ -456,6 +517,42 @@ Keyboard* keyboard_new()
     return G_KEYBOARD(pObject);
 }
 
+static void onUserChanged (GDBusConnection *pConnection, const gchar *sSender, const gchar *sPath, const gchar *sInterface, const gchar *sSignal, GVariant *pParameters, gpointer pUserData)
+{
+    Keyboard *self = G_KEYBOARD (pUserData);
+    g_variant_get (pParameters, "(s)", &self->pPrivate->sUser);
+    ActUserManager *pManager = act_user_manager_get_default ();
+    gboolean bIsLoaded;
+    g_object_get (pManager, "is-loaded", &bIsLoaded, NULL);
+
+    if (bIsLoaded)
+    {
+        if (!self->pPrivate->lUsers)
+        {
+            self->pPrivate->lUsers = act_user_manager_list_users(pManager);
+        }
+
+        for (GSList *lUser = self->pPrivate->lUsers; lUser; lUser = lUser->next)
+        {
+            ActUser *pUser = lUser->data;
+            gboolean bIsUserLoaded = act_user_is_loaded (pUser);
+
+            if (bIsUserLoaded)
+            {
+                getAccountsService (self, pUser);
+            }
+            else
+            {
+                g_signal_connect_swapped (pUser, "notify::is-loaded", G_CALLBACK (onUserLoaded), self);
+            }
+        }
+    }
+    else
+    {
+        g_signal_connect_object (pManager, "notify::is-loaded", G_CALLBACK (onManagerLoaded), self, G_CONNECT_SWAPPED);
+    }
+}
+
 static void keyboard_init(Keyboard *self)
 {
     self->pPrivate = keyboard_get_instance_private(self);
@@ -499,13 +596,31 @@ static void keyboard_init(Keyboard *self)
             pLayout->sDescription = g_strdup(sDescription);
         }
 
-        g_hash_table_replace(self->pPrivate->lLayouts, pLayout->sId, pLayout);
+        gboolean bContains = g_hash_table_contains (self->pPrivate->lLayouts, pLayout->sId);
 
+        if (!bContains)
+        {
+            g_debug ("    {\"%s\", \"%s\"}, //%s", pLayout->sLanguage, pLayout->sId, pLayout->sDescription);
+        }
+
+        g_hash_table_replace(self->pPrivate->lLayouts, pLayout->sId, pLayout);
         pRxkbLayout = rxkb_layout_next(pRxkbLayout);
     }
 
     rxkb_context_unref(pContext);
 
+    GError *pError = NULL;
+    self->pPrivate->sUser = NULL;
+    self->pPrivate->pConnection = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &pError);
+
+    if (pError)
+    {
+        g_error ("Panic: Failed connecting to the session bus: %s", pError->message);
+        g_error_free (pError);
+
+        return;
+    }
+
     if (isGreeter() == FALSE)
     {
         self->pPrivate->pDisplay = XOpenDisplay(NULL);
@@ -527,107 +642,148 @@ static void keyboard_init(Keyboard *self)
     }
     else
     {
-        // Get layouts from /etc/default/keyboard
-        gchar *sFile;
-        g_file_get_contents ("/etc/default/keyboard", &sFile, NULL, NULL);
-        gchar **lLines = g_strsplit(sFile, "\n", -1);
-        guint nLines = g_strv_length(lLines);
-        gchar **lLayouts = NULL;
-        gchar **lVariants = NULL;
+        self->pPrivate->lUsers = NULL;
+        self->pPrivate->nSubscription = g_dbus_connection_signal_subscribe (self->pPrivate->pConnection, NULL, GREETER_BUS_NAME, "UserChanged", GREETER_BUS_PATH, NULL, G_DBUS_SIGNAL_FLAGS_NONE, onUserChanged, self, NULL);
 
-        for (guint nLine = 0; nLine < nLines; nLine++)
+        // Get system layout
+        gboolean bDefaultLocation = g_file_test ("/etc/default/keyboard", G_FILE_TEST_EXISTS);
+        gchar *sLocation = NULL;
+
+        if (bDefaultLocation)
         {
-            gboolean bIsLayout = g_str_has_prefix(lLines[nLine], "XKBLAYOUT");
-
-            if (bIsLayout == TRUE)
-            {
-                gboolean bQuoted = g_strrstr(lLines[nLine], "\"") != NULL;
-                gchar *sLayout = NULL;
-
-                if (bQuoted == TRUE)
-                {
-                    sLayout = (lLines[nLine] + 11);
-                    guint nLength = strlen(sLayout);
-                    sLayout[nLength - 1] = '\0';
-                }
-                else
-                {
-                    sLayout = (lLines[nLine] + 10);
-                }
-
-                lLayouts = g_strsplit(sLayout, ",", -1);
-
-                continue;
-            }
-
-            gboolean bIsVariant = g_str_has_prefix(lLines[nLine], "XKBVARIANT");
-
-            if (bIsVariant == TRUE)
-            {
-                gboolean bQuoted = g_strrstr(lLines[nLine], "\"") != NULL;
-                gchar *sVariant = NULL;
-
-                if (bQuoted == TRUE)
-                {
-                    sVariant = (lLines[nLine] + 12);
-                    guint nLength = strlen(sVariant);
-                    sVariant[nLength - 1] = '\0';
-                }
-                else
-                {
-                    sVariant = (lLines[nLine] + 11);
-                }
-
-                lVariants = g_strsplit(sVariant, ",", -1);
-
-                continue;
-            }
+            sLocation = "/etc/default/keyboard";
+        }
+        else
+        {
+            sLocation = "/etc/X11/xorg.conf.d/00-keyboard.conf";
         }
 
-        guint nLayouts = g_strv_length(lLayouts);
-        guint nVariants = 0;
+        gchar *sFile = NULL;
+        GError *pError = NULL;
+        g_file_get_contents (sLocation, &sFile, NULL, &pError);
+        gchar *sLayout = NULL;
+        gchar *sVariant = NULL;
 
-        if (lVariants != NULL)
+        if (!pError)
         {
-            g_strv_length(lVariants);
-        }
+            GRegex *pRegex = NULL;
 
-        for (guint nLayout = 0; nLayout < nLayouts; nLayout++)
-        {
-            gchar *sId = NULL;
-
-            if (nVariants > nLayout)
+            if (bDefaultLocation)
             {
-                guint nVariant = strlen(lVariants[nLayout]);
-
-                if (nVariants == nLayouts && nVariant > 0)
-                {
-                    sId = g_strconcat(lLayouts[nLayout], "+", lVariants[nLayout], NULL);
-                }
-                else
-                {
-                    sId = g_strdup(lLayouts[nLayout]);
-                }
+                #if GLIB_CHECK_VERSION(2, 73, 0)
+                    pRegex = g_regex_new (" *XKBLAYOUT *= *\"*([a-zA-Z]*)\"*", G_REGEX_DEFAULT, G_REGEX_MATCH_DEFAULT, &pError);
+                #else
+                    pRegex = g_regex_new (" *XKBLAYOUT *= *\"*([a-zA-Z]*)\"*", (GRegexCompileFlags) 0, (GRegexMatchFlags) 0, &pError);
+                #endif
             }
             else
             {
-                sId = g_strdup(lLayouts[nLayout]);
+                #if GLIB_CHECK_VERSION(2, 73, 0)
+                    pRegex = g_regex_new (" *Option +\"*XkbLayout\"* +\"*([a-zA-Z]*)\"*", G_REGEX_DEFAULT, G_REGEX_MATCH_DEFAULT, &pError);
+                #else
+                    pRegex = g_regex_new (" *Option +\"*XkbLayout\"* +\"*([a-zA-Z]*)\"*", (GRegexCompileFlags) 0, (GRegexMatchFlags) 0, &pError);
+                #endif
             }
 
-            self->pPrivate->lLayoutRec = g_slist_append(self->pPrivate->lLayoutRec, sId);
+            if (!pError)
+            {
+                GMatchInfo *pMatchInfo = NULL;
+
+                #if GLIB_CHECK_VERSION(2, 73, 0)
+                    gboolean bMatch = g_regex_match (pRegex, sFile, G_REGEX_MATCH_DEFAULT, &pMatchInfo);
+                #else
+                    gboolean bMatch = g_regex_match (pRegex, sFile, (GRegexMatchFlags) 0, &pMatchInfo);
+                #endif
+
+                if (bMatch)
+                {
+                    sLayout = g_match_info_fetch (pMatchInfo, 1);
+                }
+                else
+                {
+                    g_error ("PANIC: No system XkbLayout found");
+                }
+
+                g_match_info_free (pMatchInfo);
+                g_regex_unref (pRegex);
+            }
+            else
+            {
+                g_error ("PANIC: Failed to compile regex: %s", pError->message);
+                g_error_free (pError);
+            }
+
+            if (bDefaultLocation)
+            {
+                #if GLIB_CHECK_VERSION(2, 73, 0)
+                    pRegex = g_regex_new (" *XKBVARIANT *= *\"*([a-zA-Z]*)\"*", G_REGEX_DEFAULT, G_REGEX_MATCH_DEFAULT, &pError);
+                #else
+                    pRegex = g_regex_new (" *XKBVARIANT *= *\"*([a-zA-Z]*)\"*", (GRegexCompileFlags) 0, (GRegexMatchFlags) 0, &pError);
+                #endif
+            }
+            else
+            {
+                #if GLIB_CHECK_VERSION(2, 73, 0)
+                    pRegex = g_regex_new (" *Option +\"*XkbVariant\"* +\"*([a-zA-Z]*)\"*", G_REGEX_DEFAULT, G_REGEX_MATCH_DEFAULT, &pError);
+                #else
+                    pRegex = g_regex_new (" *Option +\"*XkbVariant\"* +\"*([a-zA-Z]*)\"*", (GRegexCompileFlags) 0, (GRegexMatchFlags) 0, &pError);
+                #endif
+            }
+
+            if (!pError)
+            {
+                GMatchInfo *pMatchInfo = NULL;
+
+                #if GLIB_CHECK_VERSION(2, 73, 0)
+                    gboolean bMatch = g_regex_match (pRegex, sFile, G_REGEX_MATCH_DEFAULT, &pMatchInfo);
+                #else
+                    gboolean bMatch = g_regex_match (pRegex, sFile, (GRegexMatchFlags) 0, &pMatchInfo);
+                #endif
+
+                if (bMatch)
+                {
+                    sVariant = g_match_info_fetch (pMatchInfo, 1);
+                }
+
+                g_match_info_free (pMatchInfo);
+                g_regex_unref (pRegex);
+            }
+            else
+            {
+                g_error ("PANIC: Failed to compile regex: %s", pError->message);
+                g_error_free (pError);
+            }
+
+            g_free(sFile);
         }
-
-        self->pPrivate->nLayout = 0;
-
-        g_strfreev(lLayouts);
-
-        if (lVariants != NULL)
+        else
         {
-            g_strfreev(lVariants);
+            g_error ("PANIC: Failed to get %s contents: %s", sLocation, pError->message);
+            g_error_free (pError);
         }
 
-        g_strfreev(lLines);
-        g_free(sFile);
+        gchar *sId = NULL;
+        guint nVariant = 0;
+
+        if (sVariant)
+        {
+            nVariant = strlen (sVariant);
+        }
+
+        if (nVariant)
+        {
+            sId = g_strconcat (sLayout, "+", sVariant, NULL);
+        }
+        else
+        {
+            sId = g_strdup (sLayout);
+        }
+
+        g_free (sLayout);
+        g_free (sVariant);
+        self->pPrivate->lLayoutRec = g_slist_append (self->pPrivate->lLayoutRec, sId);
+        self->pPrivate->sSystemLayout = g_strdup (sId);
+        self->pPrivate->nLayout = 0;
 
         ActUserManager *pManager = act_user_manager_get_default();
         gboolean bIsLoaded;
@@ -644,7 +800,7 @@ static void keyboard_init(Keyboard *self)
 
                 if (bIsUserLoaded)
                 {
-                    getAccountsService(self);
+                    getAccountsService(self, pUser);
                 }
                 else
                 {
